/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.schemes;

import jalview.analysis.scoremodels.PairwiseSeqScoreModel;
import jalview.math.Matrix;
import jalview.math.MatrixI;

public class ScoreMatrix extends PairwiseSeqScoreModel
{
  String name;

  @Override
  public String getName()
  {
    return name;
  }

  /**
   * reference to integer score matrix
   */
  int[][] matrix;

  /**
   * 0 for Protein Score matrix. 1 for dna score matrix
   */
  int type;

  /**
   * 
   * @param name
   *          Unique, human readable name for the matrix
   * @param matrix
   *          Pairwise scores indexed according to appropriate symbol alphabet
   * @param type
   *          0 for Protein, 1 for NA
   */
  ScoreMatrix(String name, int[][] matrix, int type)
  {
    this.matrix = matrix;
    this.type = type;
    this.name = name;
  }

  @Override
  public boolean isDNA()
  {
    return type == 1;
  }

  @Override
  public boolean isProtein()
  {
    return type == 0;
  }

  @Override
  public int[][] getMatrix()
  {
    return matrix;
  }

  /**
   * Answers the score for substituting first char in A1 with first char in A2
   * 
   * @param A1
   * @param A2
   * @return
   */
  public int getPairwiseScore(String A1, String A2)
  {
    return getPairwiseScore(A1.charAt(0), A2.charAt(0));
  }

  @Override
  public int getPairwiseScore(char c, char d)
  {
    int pog = 0;

    try
    {
      int a = (type == 0) ? ResidueProperties.aaIndex[c]
              : ResidueProperties.nucleotideIndex[c];
      int b = (type == 0) ? ResidueProperties.aaIndex[d]
              : ResidueProperties.nucleotideIndex[d];

      /*
       * FIXME: 2.10.1 PCA treats gap as [22] or 'X', but Tree
       * calculation treats as [23]; which is correct?
       */
      /*
       * hack to convert unassigned / unknown (including gap)
       * to index of unknown (X for amino acids, N for nucleotide)
       * TODO: statically assign gap characters to this index?
       */
//      if (type == 0)
//      {
//        if (a == ResidueProperties.maxProteinIndex)
//        {
//          a = ResidueProperties.aaIndex['X'];
//        }
//        if (b == ResidueProperties.maxProteinIndex)
//        {
//          b = ResidueProperties.aaIndex['X'];
//        }
//      }
//      if (type != 0)
//      {
//        if (a == ResidueProperties.maxNucleotideIndex)
//        {
//          a = ResidueProperties.nucleotideIndex['N'];
//        }
//        if (b == ResidueProperties.maxNucleotideIndex)
//        {
//          b = ResidueProperties.nucleotideIndex['N'];
//        }
//      }
      pog = matrix[a][b];
    } catch (Exception e)
    {
      // System.out.println("Unknown residue in " + A1 + " " + A2);
    }

    return pog;
  }

  /**
   * pretty print the matrix
   */
  @Override
  public String toString()
  {
    return outputMatrix(false);
  }

  public String outputMatrix(boolean html)
  {
    StringBuffer sb = new StringBuffer();
    int[] symbols = (type == 0) ? ResidueProperties.aaIndex
            : ResidueProperties.nucleotideIndex;
    int symMax = (type == 0) ? ResidueProperties.maxProteinIndex
            : ResidueProperties.maxNucleotideIndex;
    boolean header = true;
    if (html)
    {
      sb.append("<table border=\"1\">");
    }
    for (char sym = 'A'; sym <= 'Z'; sym++)
    {
      if (symbols[sym] >= 0 && symbols[sym] < symMax)
      {
        if (header)
        {
          sb.append(html ? "<tr><td></td>" : "");
          for (char sym2 = 'A'; sym2 <= 'Z'; sym2++)
          {
            if (symbols[sym2] >= 0 && symbols[sym2] < symMax)
            {
              sb.append((html ? "<td>&nbsp;" : "\t") + sym2
                      + (html ? "&nbsp;</td>" : ""));
            }
          }
          header = false;
          sb.append(html ? "</tr>\n" : "\n");
        }
        if (html)
        {
          sb.append("<tr>");
        }
        sb.append((html ? "<td>" : "") + sym + (html ? "</td>" : ""));
        for (char sym2 = 'A'; sym2 <= 'Z'; sym2++)
        {
          if (symbols[sym2] >= 0 && symbols[sym2] < symMax)
          {
            sb.append((html ? "<td>" : "\t")
                    + matrix[symbols[sym]][symbols[sym2]]
                    + (html ? "</td>" : ""));
          }
        }
        sb.append(html ? "</tr>\n" : "\n");
      }
    }
    if (html)
    {
      sb.append("</table>");
    }
    return sb.toString();
  }

  /**
   * Computes an NxN matrix where N is the number of sequences, and entry [i, j]
   * is sequence[i] pairwise multiplied with sequence[j], as a sum of scores
   * computed using the current score matrix. For example
   * <ul>
   * <li>Sequences:</li>
   * <li>FKL</li>
   * <li>R-D</li>
   * <li>QIA</li>
   * <li>GWC</li>
   * <li>Score matrix is BLOSUM62</li>
   * <li>Gaps treated same as X (unknown)</li>
   * <li>product [0, 0] = F.F + K.K + L.L = 6 + 5 + 4 = 15</li>
   * <li>product [1, 1] = R.R + -.- + D.D = 5 + -1 + 6 = 10</li>
   * <li>product [2, 2] = Q.Q + I.I + A.A = 5 + 4 + 4 = 13</li>
   * <li>product [3, 3] = G.G + W.W + C.C = 6 + 11 + 9 = 26</li>
   * <li>product[0, 1] = F.R + K.- + L.D = -3 + -1 + -3 = -8
   * <li>and so on</li>
   * </ul>
   */
  public MatrixI computePairwiseScores(String[] seqs)
  {
    double[][] values = new double[seqs.length][];
    for (int row = 0; row < seqs.length; row++)
    {
      values[row] = new double[seqs.length];
      for (int col = 0; col < seqs.length; col++)
      {
        int total = 0;
        int width = Math.min(seqs[row].length(), seqs[col].length());
        for (int i = 0; i < width; i++)
        {
          char c1 = seqs[row].charAt(i);
          char c2 = seqs[col].charAt(i);
          int score = getPairwiseScore(c1, c2);
          total += score;
        }
        values[row][col] = total;
      }
    }
    return new Matrix(values);
  }
}
